/*
 * Copyright (C) 2011 Simon A. Berger
 * Portions Copyright (c) 2011 Nikolaos Alachiotis
 *
 *  This program is free software; you may redistribute it and/or modify its
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */

#include <vector>
#include <map>
#include <algorithm>
#include <numeric>
#include <iterator>
#include <iostream>
#include <fstream>
#include <memory>
#include <stdexcept>
#include <stdint.h>
#include <cassert>

#include <sys/time.h>

#include <CL/opencl.h>
#include "setupGPU.h"
#include "memCNTRL.h"
#include "utilsPaPaRa.h"
#include "align_pvec_vec.h"
#include "papara_gpu.h"
#include "../ivy_mike/src/ivymike/time.h"
static double gettime(void )
{
    struct timeval ttime;
    gettimeofday(&ttime , 0);
    return ttime.tv_sec + ttime.tv_usec * 0.000001;
}


class timer {
    double m_st;

public:
    timer() : m_st(gettime()) {};

    double elapsed() const {
        return gettime() - m_st;
    }


};



class non_copy {
	non_copy( const non_copy & ) {}
	const non_copy &operator=(const non_copy & ) { return *this; }
public:
	non_copy() {}
};

class papara_opencl : public non_copy {
	const char * kernel_code;
	cl_program myProgram;
	cl_kernel myKernel;
	cl_context myContext;
	cl_command_queue myCommandQueue;

	size_t globalsizeINTS;

	size_t BLOCKlengthCOLUMNS;
	size_t shared_memory_size;//KBs
	size_t templineSMaxLen;


	// QS
	cl_mem m_QueVecDev;
	size_t quelen_max;
	size_t que_num;
	std::vector<int> qs_len;

	cl_mem m_SLSCVecDev;
	cl_mem m_BLOCKAW_lineDev;

	// ref
	const size_t ref_len;
	const size_t workgroup_width;
	cl_mem m_InterRefVecDev_a;
	cl_mem m_ArgVecDev;


	// output
	cl_mem m_ResVecPin;
	cl_mem m_ResVecDev;

	//std::map<std::pair<size_t, size_t>, int> m_res_map;
	//std::vector<int> m_res;
	size_t m_cups;

	size_t size_query_mem_;
	size_t size_ref_mem_;

//	papara_opencl( const papara_opencl & ) {}; // non-copy
//	const papara_opencl &operator=( const papara_opencl & ) { return *this; } // non-copy
public:
	papara_opencl( size_t ref_len_, size_t workgroup_width_ )
	: globalsizeINTS(0),
	  m_QueVecDev(0), quelen_max(0), que_num(0),
	  m_SLSCVecDev(0), m_BLOCKAW_lineDev(0),

	  ref_len(ref_len_), workgroup_width(workgroup_width_),
	  m_InterRefVecDev_a(0), m_ArgVecDev(0),
	  m_ResVecPin(0), m_ResVecDev(0),
	  m_cups(0),
	  size_query_mem_(0), size_ref_mem_(0)


	{




	   	cl_platform_id myPlatform = select_NVIDIA_platform();

		cl_device_id  myDevice = select_GPU_device(myPlatform);

		myContext = create_context(myDevice);

		myCommandQueue = create_command_queue (myDevice, myContext);

		int kernel_size;



		/* Generic Kernel *
		char * kernel_code = load_kernel_code ("PaPaRa-OpenCL-Generic-Kernel.cl", &kernel_size);
		cl_program myProgram = create_and_build_program(myDevice, myContext, kernel_code, kernel_size);
		cl_kernel myKernel = create_kernel(myProgram, "PaPaRaOpenCLGenericKernel");
		*/

		/* Compressed Kernel 12*/
		std::string kernel_code = load_kernel_code ("PaPaRa-OpenCL-Unfold12-Kernel-Compressed.cl" );


		myProgram = create_and_build_program(myDevice, myContext, kernel_code.c_str(), kernel_code.size() );
		myKernel = create_kernel(myProgram, "PaPaRaOpenCLUnfold12KernelCompressed");


		// Shared Memory
//		shared_memory_size = 12;//KBs
		shared_memory_size = 15;
		templineSMaxLen = (shared_memory_size * 1024)/4; // integers




		// Blocked Version Parameters
		BLOCKlengthCOLUMNS = templineSMaxLen / workgroup_width;
		std::cout<<"Block size in columns:\t\t"<<BLOCKlengthCOLUMNS<<"\n";
	}


	void free_query_memory() {
		release_mem_object( m_QueVecDev );
		release_mem_object( m_SLSCVecDev );
		release_mem_object( m_BLOCKAW_lineDev );
		release_mem_object( m_ResVecDev );
		release_mem_object( m_ResVecPin );

		m_QueVecDev = 0;
		m_SLSCVecDev = 0;
		m_BLOCKAW_lineDev = 0;
		m_ResVecDev = 0;
		m_ResVecPin = 0;

		size_query_mem_ = 0;
	}

	void free_ref_memory() {
		release_mem_object(m_InterRefVecDev_a);
		release_mem_object(m_ArgVecDev);

		m_ArgVecDev = 0;
		m_InterRefVecDev_a = 0;
		size_ref_mem_ = 0;
	}


	~papara_opencl() {
		// TODO: free opencl resources

		free_query_memory();
		free_ref_memory();

		cl_uint ref = 0;
		clGetContextInfo(myContext, CL_CONTEXT_REFERENCE_COUNT, sizeof(cl_uint), &ref, 0 );

		std::cout << "context refcount: " << ref << "\n";

		clReleaseKernel(myKernel);
		clReleaseProgram(myProgram);
		clReleaseCommandQueue(myCommandQueue);
		clReleaseContext(myContext);



	}


	void setup_qs( const std::vector< const std::vector<uint8_t> *> & qs ) {
		free_query_memory();



		que_num = qs.size();


		//		for(size_t i=0;i<qs.size();i++) {
		//			//quelen_max = std::max( quelen_max, qs[i].size() );
		//			qs_len.push_back( qs[i]->size() );
		//		}

		qs_len.resize(qs.size());
		std::transform(qs.begin(), qs.end(), qs_len.begin(), std::mem_fun(&std::vector<uint8_t>::size)); // <- every time you use a for-loop instead of an stl-algorithm, a baby-seal will die (according to Scott Meyers)


		assert( ! qs_len.empty() );

		quelen_max = *std::max_element( qs_len.begin(), qs_len.end() );

		//std::cout << "Number of References:\t\t"<< RefNum << "\nNumber of Queries:\t\t" << QueNum << "\nReference Length:\t\t" << RefLen << "\nMaximum Query Length:\t\t" << QueLenMax <<"\n";
		std::cout << "Number of Queries:\t\t" << qs.size() << "\nMaximum Query Length:\t\t" << quelen_max <<"\n";




		// Query Organization: Placement one after the other in blocks of size QueLenMax and zero filling.
		const size_t QueVecLen = qs.size() * quelen_max; // number of unsigned integers in the QueVec

		//unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
		std::vector<unsigned int> que_vec( QueVecLen );

		for( size_t i=0; i<qs.size(); i++){
			putSequenceQue( i, *qs[i], quelen_max, que_vec.data());
		}


		cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
		QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, que_vec.data());
		m_QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
		copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, m_QueVecDev, myCommandQueue);


		size_query_mem_ += QueVecLen * sizeof(uint32_t);
		globalsizeINTS += QueVecLen;

		const size_t ResVecLen = que_num * workgroup_width;

		m_ResVecPin = allocatePinnedHostMemory_32 (ResVecLen, myContext, myCommandQueue);
		m_ResVecDev = allocateDeviceMemory_32 (ResVecLen, myContext);

		size_query_mem_ += ResVecLen * sizeof(uint32_t);

		const size_t tempcolGlen  = que_num * quelen_max * workgroup_width;
		const size_t templineGlen = que_num * (quelen_max + BLOCKlengthCOLUMNS) * workgroup_width;
		//const size_t templineSlen = que_num * BLOCKlengthCOLUMNS * workgroup_width;



		//cl_mem SLVecDev = allocateDeviceMemory_32 (tempcolGlen , myContext);
		//cl_mem SCVecDev = allocateDeviceMemory_32 (tempcolGlen , myContext);
		m_SLSCVecDev = allocateDeviceMemory_u32 (tempcolGlen, myContext);
		globalsizeINTS+=tempcolGlen;
		m_BLOCKAW_lineDev = allocateDeviceMemory_32 (templineGlen , myContext);
		globalsizeINTS+=templineGlen;
		//cl_mem BLOCK_lineDev = allocateDeviceMemory_32 (templineSlen, myContext);

		size_query_mem_ += tempcolGlen * sizeof(uint32_t);
		size_query_mem_ += templineGlen * sizeof(uint32_t);

		clFinish(myCommandQueue);

		release_mem_object(QueVecPin);
	}


	std::pair<size_t,size_t> calculate_qs_size( size_t num, size_t max ) {
		const size_t QueVecLen = num * max * sizeof(uint32_t);
		const size_t ResVecLen = num * workgroup_width * sizeof(uint32_t);
		const size_t tempcolGlen  = num * max * workgroup_width * sizeof(uint32_t);
		const size_t templineGlen = num * (max + BLOCKlengthCOLUMNS) * workgroup_width * sizeof(uint32_t);

		return std::make_pair(QueVecLen + ResVecLen + tempcolGlen + templineGlen, std::max( std::max(QueVecLen, ResVecLen), std::max(tempcolGlen, templineGlen )) );
	}

	std::vector<int> run_refs( std::vector<const std::vector<int> *> & ref_pvecs, std::vector<const std::vector<unsigned int> *> & ref_aux ) {
		// Inter-Reference Memory Organization

		ivy_mike::perf_timer pt1;
		free_ref_memory();

		std::cerr << workgroup_width << " " << get_next32mult(workgroup_width) << "\n";
		assert( workgroup_width == get_next32mult(workgroup_width));


		//const size_t RefPadBloLen = get_next32mult ( workgroup_width );
		std::cout << "Reference Block Size:\t\t"<<workgroup_width<<"\n\n";
//		const size_t InterRefVecLen = workgroup_width * ref_len;
		//const float InterRefVecGBs = ((float)InterRefVecLen*32.0)/8.0/1024.0/1024.0/1024.0;

		//int * InterRefVec_a = (int *) malloc(sizeof(int)*InterRefVecLen);
		//unsigned int * InterRefVec_a_aux = (unsigned int *) malloc(sizeof(unsigned int)*InterRefVecLen);
		//for(size_t i=0;i<RefNum;i++){putSequenceRef_InterInt(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec_a,RefPadBloLen);}

	//	std::cout << "Reference Block GBs:\t\t" << InterRefVecGBs * 2 << "GBs\n\n";


//





		int total = ref_len;
		int partF = quelen_max;
		int partL = quelen_max;
		int partM = total - partF - partL;
		int blockiterations = partM / BLOCKlengthCOLUMNS;
		std::cout<<"Iterations:\t\t"<<blockiterations<<"\n";
		int sitesinblocks = blockiterations * BLOCKlengthCOLUMNS;
		int partLext = partM - sitesinblocks;
		    partL = partL + partLext;
		int lastblock_refstart = total - partL;

	//	std::cout<<"Limit1:"<< partF <<"-- Limit2:"<<lastblock_refstart<<"LastL:"<<partL<<"Check:"<<partF+blockiterations*BLOCKlengthCOLUMNS+partL;

		// Inter-Reference Memory Middle Part Compression: 4 ref entries per int
		const size_t compression_effort = 6;
		const size_t compressed_vector_length = ((lastblock_refstart - partF)/compression_effort)*workgroup_width;
	//	std::cout<<"Compressed Vector Length:"<<compressed_vector_length;

		//int * CompRefMidVec = (int *) malloc(sizeof(int)*compressed_vector_length);
		std::vector<int> CompRefMidVec( compressed_vector_length );

		for(size_t i=0; i < workgroup_width; i++) {
			putSequenceRef_InterIntComp(i, *(ref_pvecs[i]), *(ref_aux[i]), ref_len, partF, lastblock_refstart,CompRefMidVec.data(), workgroup_width);
		}


		// Merge uncompressed and compressed values

		const size_t RefVecLen_Combined = (partF + partL + blockiterations*BLOCKlengthCOLUMNS/compression_effort) * workgroup_width;
		std::cout<<"Combined length:"<<RefVecLen_Combined;
		//int * RefVecCombined = (int *) malloc (sizeof(int)*RefVecLen_Combined);
		std::vector<int> RefVecCombined( RefVecLen_Combined );


		for(size_t i=0;i< workgroup_width;i++) {
			putSequenceRef_CombinedS1(i, *(ref_pvecs[i]), *(ref_aux[i]), partF, RefVecCombined.data(), workgroup_width);
		}

	//	std::cout<<"1:"<< partF <<"\n2:"<<blockiterations<<"\n3:"<<sitesinblocks<<"\n4:"<<lastblock_refstart;

		for(size_t i=0;i<workgroup_width;i++) {
			putSequenceRef_CombinedS2(i, *(ref_pvecs[i]), *(ref_aux[i]), partL, RefVecCombined.data(), workgroup_width, partF, blockiterations, BLOCKlengthCOLUMNS/compression_effort,lastblock_refstart);
		}

		for(size_t i=0;i < workgroup_width; i++ ) {
			putSequenceRef_InterIntCompCombined(i, *(ref_pvecs[i]), *(ref_aux[i]), workgroup_width,partF, lastblock_refstart,RefVecCombined.data(),workgroup_width,partF);
		}

		//unsigned int quelenmax, unsigned int blockiterations, unsigned int blocksizecompressed, unsigned int startpos);

		pt1.add_int();
		// Arguments
		const size_t ArgVecLen = que_num + 9; // Set to RefLen for kernel testing
		//int * ArgVec = (int *) malloc(sizeof(int)*ArgVecLen);

		std::vector<int> ArgVec( ArgVecLen );

		ArgVec[0] = que_num;
		ArgVec[1] = workgroup_width;
		ArgVec[2] = ref_len;
		ArgVec[3] = workgroup_width;
		ArgVec[4] = quelen_max;
		ArgVec[5] = BLOCKlengthCOLUMNS;
		ArgVec[6] = blockiterations;
		ArgVec[7] = lastblock_refstart;
		ArgVec[8] = lastblock_refstart - (quelen_max + blockiterations * BLOCKlengthCOLUMNS / compression_effort);

	//	std::cout<< "previous:"<<ArgVec[7]<<"-current:"<<ArgVec[8]<<"\n";

		for(size_t i = 9; i< ArgVecLen; i++)
		{
			ArgVec[i] = qs_len[i-9];
		}
		//const size_t InterIntVecLen = que_num *  InterRefVecLen;


		// GPU Memory Allocation


		assert( m_InterRefVecDev_a == 0 );
		cl_mem InterRefVecPin_a = allocatePinnedHostMemory_32 (RefVecLen_Combined, myContext, myCommandQueue);
		InterRefVecPin_a = initializePinnedHostMemory_32_Sequence (RefVecLen_Combined, InterRefVecPin_a, myCommandQueue, RefVecCombined.data());
		m_InterRefVecDev_a = allocateDeviceMemory_32 (RefVecLen_Combined, myContext);
		copyPinnedHostMemoryToDeviceMemory_32 (RefVecLen_Combined, InterRefVecPin_a, m_InterRefVecDev_a, myCommandQueue );

		size_ref_mem_ += RefVecLen_Combined * sizeof(uint32_t);

		globalsizeINTS+=RefVecLen_Combined;

		assert( m_ArgVecDev == 0 );
		cl_mem ArgVecPin = allocatePinnedHostMemory_32 (ArgVecLen, myContext, myCommandQueue);
		ArgVecPin = initializePinnedHostMemory_32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec.data());
		m_ArgVecDev = allocateDeviceMemory_32 (ArgVecLen, myContext);
		copyPinnedHostMemoryToDeviceMemory_32 (ArgVecLen, ArgVecPin, m_ArgVecDev, myCommandQueue );

		globalsizeINTS+=ArgVecLen;
		size_ref_mem_ += ArgVecLen * sizeof(uint32_t);


//		globalsizeINTS+=ResVecLen;

		//clFinish(myCommandQueue);
		release_mem_object(InterRefVecPin_a);
		release_mem_object(ArgVecPin);


//		clGetKernelInfo()
		// Launch GPU kernel

		cl_int ret;
		ret = clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &m_QueVecDev); ocl_check(ret, CL_SUCCESS);
		ret = clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &m_InterRefVecDev_a); ocl_check(ret, CL_SUCCESS);
		ret = clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &m_ArgVecDev); ocl_check(ret, CL_SUCCESS);
		ret = clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &m_ResVecDev); ocl_check(ret, CL_SUCCESS);

		//clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &SLVecDev);
		//clSetKernelArg(myKernel, 5, sizeof(cl_mem), (void *) &SCVecDev);
		ret = clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &m_BLOCKAW_lineDev); ocl_check(ret, CL_SUCCESS);
		ret = clSetKernelArg(myKernel, 5, sizeof(int) * templineSMaxLen, 0); ocl_check(ret, CL_SUCCESS);
		ret = clSetKernelArg(myKernel, 6, sizeof(cl_mem), (void *) &m_SLSCVecDev); ocl_check(ret, CL_SUCCESS);
		//clSetKernelArg(myKernel, 7, sizeof(cl_mem), (void *) &InterRefVecCompDev_a);
		//cl_int fret = clFinish(myCommandQueue);

		pt1.add_int();

//		assert( fret == CL_SUCCESS );

		size_t WorkSizeGlobal[] = {workgroup_width * que_num};
		size_t WorkSizeLocal[] = {workgroup_width};

		//cl_event event;
		timer NikosTimerN;


		cl_int fret = clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, 0);
		if( fret != CL_SUCCESS ) {
			std::cerr << "err: " << fret << "\n";
		}

		ocl_check( fret, CL_SUCCESS );

//		std::cout << "kernel started\n";
		while( NikosTimerN.elapsed() < 5 ) {
			usleep(20000);
		}

//		std::cout << "finish\n";
		//fret = clFinish(myCommandQueue);




		fret = clFinish(myCommandQueue);

		ocl_check( fret, CL_SUCCESS );

		std::cout << "time: " << NikosTimerN.elapsed() << "\n";

		const size_t ResVecLen = que_num * workgroup_width;
		int * ResVec = copyDeviceMemoryToPinnedHostMemory_32 (ResVecLen, m_ResVecDev, m_ResVecPin, myCommandQueue, false);
		clFinish(myCommandQueue);

		//m_res.resize(ResVecLen);

		std::vector<int> res(ResVecLen);
		std::copy( ResVec, ResVec + ResVecLen, res.begin() );

		clEnqueueUnmapMemObject( myCommandQueue, m_ResVecPin, ResVec, 0, 0, 0 );

		//size_t qs_all_len = std::accumulate( qs_len.begin(), qs_len.end(), 0 );

		double qs_mean_len = std::accumulate( qs_len.begin(), qs_len.end(), 0 ) / double(qs_len.size());

		m_cups = (ref_pvecs.front()->size() - qs_mean_len) * qs_mean_len * qs_len.size() * ref_pvecs.size(); // <- maybe I'm too stupid, but this term can surely be simplified...

		std::cout << "device mem: " << size_ref_mem_ * 1e-6 << " " << size_query_mem_ * 1e-6 << "\n";

		//m_cups = ((ref_pvecs.front()->size() * qs_len.size()) - qs_all_len) * ref_pvecs.size();

//		m_cups = (ref_pvecs.front()->size() - 100) * 100 * ref_pvecs.size() * qs_len.size();
//		std::copy( ResVec, ResVec + ResVecLen, std::ostream_iterator<int>(std::cout, " " ) );

//		int * res_iter = ResVec;
//		//m_res_map.clear();
//		for( size_t i = 0; i < que_num; ++i ) {
//			for( size_t j = 0; j < ref_pvecs.size(); ++j ) {
//				assert( res_iter < ResVec + ResVecLen );
//
//				m_res_map[std::make_pair(j,i)] = *res_iter;
//
//				res_iter++;
//			}
//		}

		//clEnqueueUnmapMemObject()
		pt1.add_int();
		pt1.print();

		free_ref_memory();
		free_query_memory();
		return res;
	}

//	const std::vector<int> &get_res() {
//		return m_res;
//	}

	size_t get_cups() {
		return m_cups;
	}

};


class papara_gpu_driver2_impl : public papara_gpu_driver2 {
	typedef papara_opencl gpu_core;

    const std::vector<std::vector <uint8_t> > &m_qs_pvecs;

    const static size_t workgroup_width = 320;
//    const static size_t workgroup_width = 256;

    //std::map< std::pair<size_t, size_t>, int> m_res;

    gpu_core m_core;

    ivy_mike::perf_timer m_pt1;

public:
    papara_gpu_driver2_impl( size_t ref_len, const std::vector<std::vector <uint8_t> > &qs )
     : m_qs_pvecs(qs), m_core( ref_len, workgroup_width )
    {}

    virtual ~papara_gpu_driver2_impl() {
    	m_pt1.print();
    }


    std::vector<int> run_impl( papara_opencl *core, const std::vector<std::vector<int> * > &refs, const std::vector<std::vector<unsigned int> * > &aux, const std::vector<size_t> &ref_ids ) {
    //	const size_t ref_len = m_ref_pvecs.front().size();
    	assert( refs.size() == workgroup_width );
    	assert( aux.size() == workgroup_width );
    	assert( ref_ids.size() == workgroup_width );


    	std::vector<int> res_all;
    	//const size_t qs_group_size = 350 * 2;

    	size_t qs_start = 0;

    	while( qs_start < m_qs_pvecs.size() ) {

    		ivy_mike::perf_timer pt;

    		std::vector<const std::vector<uint8_t> *> qs_group_pvecs;

    		size_t maxlen = 0;

    		size_t qs_end = qs_start;
    		for( size_t i = qs_start; i < m_qs_pvecs.size(); ++i ) {
    			maxlen = std::max( m_qs_pvecs[i].size(), maxlen );

    			std::pair<size_t,size_t> sizes = core->calculate_qs_size(i-qs_start + 1, maxlen);


//    			std::cout << "sizes: " << sizes.first << " " << sizes.second << "\n";
    			if( sizes.first > 700 * 1024 * 1024 || sizes.second > 250 * 1000 * 1000 ) {

    				break;
    			}

    			qs_end = i + 1;
    		}

//    		qs_end -= qs_end % 7;
//
//    		assert( qs_end > qs_start );

    		//size_t qs_end = std::min( m_qs_pvecs.size(), qs_start + qs_group_size );

    		for( size_t i = qs_start; i < qs_end; ++i ) {
    			qs_group_pvecs.push_back( &m_qs_pvecs[i] );
    		}

    		std::cout << "qs_group: " << qs_start << " - " << qs_end << "\n";

			//gpu_core core( ref_len, workgroup_width );

			core->setup_qs( qs_group_pvecs );

			pt.add_int();

			size_t ref_start = 0;

			std::vector<const std::vector <int> *> wg_pvecs;
			std::vector<const std::vector <unsigned int> *> wg_aux;

			wg_pvecs.reserve(workgroup_width);
			wg_aux.reserve(workgroup_width);

			while( ref_start < refs.size() ) {
				timer t1;

				wg_pvecs.clear();
				wg_aux.clear();

				size_t wg_valid_width = 0;

				for( size_t i = 0; i < workgroup_width; ++i ) {
					const size_t ref_idx = ref_start + i;

					if( ref_idx < refs.size() ) {
						wg_pvecs.push_back( refs[ref_idx] );
						wg_aux.push_back( aux[ref_idx] );

						wg_valid_width++;
					} else {

						assert(false);
						wg_pvecs.push_back(refs[ref_start]);
						wg_aux.push_back(aux[ref_start]);
					}
				}


				std::cout << "group offset: " << ref_start << " " << wg_valid_width << "\n";

				const std::vector<int> &res = core->run_refs(wg_pvecs, wg_aux);
				res_all.insert( res_all.end(), res.begin(), res.end() );

				//const std::vector<int> &res = core->get_res();
//				std::vector<int>::const_iterator res_iter = res.begin();
//
//				for( size_t i = 0; i < qs_group_pvecs.size(); ++i ) {
//					for( size_t j = 0; j < wg_valid_width; ++j ) {
//						assert( res_iter < res.end() );
//
//						m_res[std::make_pair(ref_ids[j],i + qs_start)] = *res_iter;
//
//						res_iter++;
//					}
//					res_iter += workgroup_width - wg_valid_width;
//				}


				ref_start += workgroup_width;

				size_t cups = core->get_cups();
				std::cout << "gncup/s: " << cups / (t1.elapsed() * 1e9) << "\n";
			}

			qs_start = qs_end;

			pt.add_int();
			m_pt1 += pt;
    	}

    	return res_all;
    }

    virtual std::vector<int> run(const std::vector< std::vector<int> *> &pvecs, const std::vector< std::vector<unsigned int> *> &aux, const std::vector<size_t> &ref_ids) {
    	//papara_opencl core( m_ref_pvecs.front().size(), workgroup_width );

    	return run_impl( &m_core, pvecs, aux, ref_ids );
    }

//    virtual const std::map<std::pair<size_t, size_t>, int> &get_res_map() {
//    	return m_res;
//    }

};



//papara_gpu_driver *new_papara_gpu_driver( const std::vector<std::vector <int> > &pvecs, const std::vector<std::vector <unsigned int> > &aux, const std::vector<std::vector <uint8_t> > &qs, size_t start, size_t end ) {
//	return new papara_gpu_driver_impl( pvecs, aux, qs, start, end );
//}

papara_gpu_driver2 *new_papara_gpu_driver2( size_t ref_len, const std::vector<std::vector <uint8_t> > &qs ) {
	return new papara_gpu_driver2_impl( ref_len, qs );
}

bool papara_gpu_driver_available() {
	return true;
}



////template<typename gpu_core>
//class papara_gpu_driver_impl : public papara_gpu_driver {
//	typedef papara_opencl gpu_core;
//
//    const std::vector<std::vector <int> > &m_ref_pvecs;
//    const std::vector<std::vector <unsigned int> > &m_ref_aux;
//
//    const std::vector<std::vector <uint8_t> > &m_qs_pvecs;
//
//    const size_t m_start;
//    const size_t m_end;
//
//    const static size_t workgroup_width = 256 ;
//
//
//    std::map< std::pair<size_t, size_t>, int> m_res;
//
//public:
//    papara_gpu_driver_impl( const std::vector<std::vector <int> > &pvecs, const std::vector<std::vector <unsigned int> > &aux, const std::vector<std::vector <uint8_t> > &qs, size_t start, size_t end )
//     : m_ref_pvecs(pvecs), m_ref_aux(aux), m_qs_pvecs(qs), m_start(start), m_end(end)
//    {}
//
//    virtual ~papara_gpu_driver_impl() {}
//
//
//    void run_impl( papara_opencl *core ) {
//    //	const size_t ref_len = m_ref_pvecs.front().size();
//
//
//    	const size_t qs_group_size = 400;
//
//    	size_t qs_start = 0;
//
//    	while( qs_start < m_qs_pvecs.size() ) {
//
//    		std::vector<const std::vector<uint8_t> *> qs_group_pvecs;
//
//    		size_t qs_end = std::min( m_qs_pvecs.size(), qs_start + qs_group_size );
//
//    		for( size_t i = qs_start; i < qs_end; ++i ) {
//    			qs_group_pvecs.push_back( &m_qs_pvecs[i] );
//    		}
//
//    		std::cout << "qs_group: " << qs_start << " - " << qs_end << "\n";
//
//			//gpu_core core( ref_len, workgroup_width );
//
//			core->setup_qs( qs_group_pvecs );
//
//			size_t ref_start = m_start;
//
//			std::vector<const std::vector <int> *> wg_pvecs;
//			std::vector<const std::vector <unsigned int> *> wg_aux;
//
//			wg_pvecs.reserve(workgroup_width);
//			wg_aux.reserve(workgroup_width);
//
//			while( ref_start < m_end ) {
//				timer t1;
//
//				wg_pvecs.clear();
//				wg_aux.clear();
//
//				size_t wg_valid_width = 0;
//
//				for( size_t i = 0; i < workgroup_width; ++i ) {
//					const size_t ref_idx = ref_start + i;
//
//					if( ref_idx < m_ref_pvecs.size() ) {
//						wg_pvecs.push_back( &m_ref_pvecs[ref_idx] );
//						wg_aux.push_back( &m_ref_aux[ref_idx] );
//
//						wg_valid_width++;
//					} else {
//						wg_pvecs.push_back(&m_ref_pvecs[ref_start]);
//						wg_aux.push_back(&m_ref_aux[ref_start]);
//					}
//				}
//
//
//				std::cout << "group offset: " << ref_start << " " << wg_valid_width << "\n";
//
//				core->run_refs(wg_pvecs, wg_aux);
//
//
//				const std::vector<int> &res = core->get_res();
//				std::vector<int>::const_iterator res_iter = res.begin();
//
//				for( size_t i = 0; i < qs_group_pvecs.size(); ++i ) {
//					for( size_t j = 0; j < wg_valid_width; ++j ) {
//						assert( res_iter < res.end() );
//
//						m_res[std::make_pair(j + ref_start,i + qs_start)] = *res_iter;
//
//						res_iter++;
//					}
//					res_iter += workgroup_width - wg_valid_width;
//				}
//
//
//				ref_start += workgroup_width;
//
//				size_t cups = core->get_cups();
//				std::cout << "gncup/s: " << cups / (t1.elapsed() * 1e9) << "\n";
//			}
//
//			qs_start = qs_end;
//    	}
//
//    }
//
//    virtual void run() {
//    	papara_opencl core( m_ref_pvecs.front().size(), workgroup_width );
//
//    	run_impl( &core );
//    }
//
//    virtual const std::map<std::pair<size_t, size_t>, int> &get_res_map() {
//    	return m_res;
//    }
//
//};



/*
    SC, SL are a_size * b_size arrays

	// remark: the SL array is not explicitly kept in my implementation because we only ever read the value that has been written in the last iteration.
	// remark: in my implementation i only keep one line of the SC array and reuse it.

    for( ib in 0..b_size-1 ) {

        astart = ib

        for( ia in astart..ib + band_width ) {
            //score_t ac = a[ia];

            bool cgap = a_aux[ia] == aux_cgap;

            // determine match or mis-match according to parsimony bits coming from the tree.
            bool is_match = ( a[ia] & b[ib] ) != 0;

			// load diagonal value from SC array
			sm = SC[ia-1,ib-1]


            if( !is_match ) {
                sm += mismatch_score;
            }


            // load left values from SL and SC, and apply gap scores
            if( cgap ) {
                last_sc_OPEN = SC[ia-1,ib];
                sl_score_stay = SL[ia-1,ib];
                sm += match_cgap;
            } else {
                last_sc_OPEN = SC[ia-1,ib] + gap_open + gap_extend;
                sl_score_stay = SL[ia-1,ib] + gap_extend;
            }



            sl = min( sl_score_stay, last_sc_OPEN );

            // write SL score for current cell
            SL[ia,ib] = sl;


			// load value from above and apply score
            su = SC[ia,ib-1] + mismatch_score;


			// find minimum of all and write current SC value
            sc = min( sm, min( su, sl ) );

            SC[ia,ib] = sc;

        }
     }

 */


//typedef int score_t;
//
//
//struct align_arrays {
//    std::vector<score_t> s;
//};
//
//
//static score_t align_pvec_score_seq( std::vector<int> &a, std::vector<unsigned int> &a_aux, const std::vector<uint8_t> &b, score_t mismatch_score, score_t match_cgap, score_t gap_open, score_t gap_extend, align_arrays &arr ) {
//
//    if( arr.s.size() < a.size()  ) {
//        arr.s.resize( a.size() );
//    }
//    assert( a.size() > b.size() );
//    const score_t aux_cgap = 0x1;
//
//    const size_t band_width = a.size() - b.size();
//    std::fill( arr.s.begin(), arr.s.end(), 0 );
//
//    const score_t LARGE = 32000;
////     std::fill( arr.si.begin(), arr.si.end(), LARGE );
//    arr.s.at(band_width+1) = LARGE;
//
////     score_t min_score = LARGE;
//
//    for( size_t ib = 0; ib < b.size(); ib++ ) {
//        int bc = b[ib];
//
//        score_t last_sl = LARGE;
//        score_t last_sc = LARGE;
//        score_t last_sdiag = 0.0;
//
//        size_t astart = ib;
//
//        score_t * __restrict s_iter = &arr.s[0];
//
//
//       // bool lastrow = ib == (b.size() - 1);
//
//
//        last_sdiag = *s_iter;
//
//
//        for( size_t ia = astart; ia <= ib + band_width; ++ia, ++s_iter ) {
//            score_t ac = a[ia];
//            const bool cgap = a_aux[ia] == (unsigned int)aux_cgap;
//
//            // determine match or mis-match according to parsimony bits coming from the tree.
//            bool is_match = ( ac & bc ) != 0;
//
//            score_t sm = last_sdiag;
//
//            if( !is_match ) {
//                sm += mismatch_score;
//            }
//
//            last_sdiag = *(s_iter+1);
//
//            score_t last_sc_OPEN;
//            score_t sl_score_stay;
//
//            if( cgap ) {
//                last_sc_OPEN = last_sc;
//                sl_score_stay = last_sl;
//                sm += match_cgap;
//            } else {
//                last_sc_OPEN = last_sc + gap_open + gap_extend;
//                sl_score_stay = last_sl + gap_extend;
//            }
//
//            score_t sl = std::min( sl_score_stay, last_sc_OPEN );
//
//            last_sl = sl;
//
//
//            score_t su = last_sdiag + mismatch_score;
//
////             std::cout << "x: " << sm << " " << su << " " << sl << "  ";
//            score_t sc = std::min( sm, std::min( su, sl ) );
//
//            last_sc = sc;
//            *s_iter = sc;
//
//
////             if( lastrow ) {
////                 if( sc < min_score ) {
////
////                     min_score = sc;
////                 }
////             }
//
//        }
//
//
//
//
////         std::cout << "\n";
////         getchar();
//    }
//
//    return *std::min_element( arr.s.begin(), arr.s.begin() + band_width + 1 );
//    //std::cout << "min2: " << min_score << " " << min2 << "\n";
//
//    //return min_score;
//
//}




//class testbench {
//    const static int score_gap_open = 1;
//    const static int score_gap_extend = 1;
//    const static int score_mismatch = 1;
//    const static int score_match_cgap = 4;
//
//    std::vector<std::vector <int> > m_ref_pvecs;
//    std::vector<std::vector <unsigned int> > m_ref_aux;
//
//    std::vector<std::vector <uint8_t> > m_qs_pvecs;
//
//    std::map<std::pair<size_t,size_t>,int> m_seq_res;
//    std::map<std::pair<size_t,size_t>,int> m_vec_res;
//    std::map<std::pair<size_t,size_t>,int> m_cl_res;
//
//    void load_qs( std::istream &is ) {
//        size_t num_qs;
//        is >> num_qs;
//
//        m_qs_pvecs.resize( num_qs );
//
//        for( size_t i = 0; i < m_qs_pvecs.size(); ++i ) {
//            size_t len;
//            is >> len;
//            while( isspace( is.get() ) && !is.eof() ) {}
//            is.unget();
//
//            if( is.eof() ) {
//                throw std::runtime_error( "unexpected end of file" );
//            }
//
//            m_qs_pvecs[i].resize(len);
//            is.read( (char*)&m_qs_pvecs[i][0], len );
////             std::cout << "qs: " << i << " " << m_qs_pvecs[i].size() << "\n";
//        }
//    }
//
//    void load_ref( std::ifstream &is ) {
//        size_t num_ref;
//        is >> num_ref;
//        m_ref_pvecs.resize( num_ref );
//        m_ref_aux.resize( num_ref );
//
//        for( size_t i = 0; i < m_ref_pvecs.size(); ++i ) {
//            size_t len;
//            is >> len;
//            while( isspace( is.get() ) && !is.eof() ) {}
//            is.unget();
//
//            if( is.eof() ) {
//                throw std::runtime_error( "unexpected end of file" );
//            }
//
//            m_ref_pvecs[i].resize(len);
//            is.read( (char*)&m_ref_pvecs[i][0], len * sizeof(int) );
//
//            m_ref_aux[i].resize(len);
//            is.read( (char*)&m_ref_aux[i][0], len * sizeof(unsigned int) );
//
////             std::cout << "ref: " << i << " " << m_ref_pvecs[i].size() << "\n";
//        }
//
//    }
//
//    template<typename T, size_t min_len>
//    static void trim_ref( std::vector<T> &v ) {
//
//    	const size_t max_len = min_len;
//        while( v.size() < min_len ) {
//            std::vector<T> v_new = v;
//
//
//            std::copy( v.begin(), v.end(), std::back_inserter(v_new) );
//
//            v.swap(v_new);
//
//        }
//
//
//        if( v.size() > max_len ) {
//            v.resize( max_len );
//        }
//
//
//    }
//    template<typename T, size_t max_len>
//    static void trim_max( std::vector<T> &v ) {
//
//        if( v.size() > max_len ) {
//            v.resize( max_len );
//        }
//
//
//    }
//
//    template<typename cont>
//	void duplicate( cont &c ) {
//    	assert( !c.empty() );
//		cont tmp = c;
//    	c.insert( c.end(), tmp.begin(), tmp.end() );
//	}
//
//public:
//    testbench( const char *qs_name, const char *ref_name, size_t num_ref, size_t num_qs ) {
//        {
//            std::ifstream is( qs_name );
//            load_qs( is );
//        }
//        {
//            std::ifstream is( ref_name );
//            load_ref( is );
//        }
//
//
//        // reduce size of testset
////        size_t num_ref = 192;
////        size_t num_qs = 200;
//
//#if 0
//        const size_t ref_len = 128;
//        const size_t qs_len = ref_len - 1;
//        std::for_each( m_ref_pvecs.begin(), m_ref_pvecs.end(), &trim_ref<int,ref_len> );
//        std::for_each( m_ref_aux.begin(), m_ref_aux.end(), &trim_ref<unsigned int,ref_len> );
//        std::for_each( m_qs_pvecs.begin(), m_qs_pvecs.end(), &trim_max<uint8_t,qs_len> );
//#endif
//
////        for( size_t i = 0; i < m_ref_pvecs.size(); ++i ) {
////        	const size_t ref_len = 4000;
////
////        	assert( m_ref_pvec)
////
////        }
//
//
//        while( m_ref_pvecs.size() < num_ref ) {
//        	duplicate( m_ref_pvecs );
//        	duplicate(m_ref_aux );
//        }
//
//        num_ref = std::min( num_ref, m_ref_pvecs.size() );
//        num_qs = std::min( num_qs, m_qs_pvecs.size() );
//
//        m_ref_pvecs.resize(num_ref);
//        m_ref_aux.resize(num_ref);
//        m_qs_pvecs.resize(num_qs);
//
//    }
//
//    void run_opencl() {
//
//    	//const size_t ref_len = m_ref_pvecs.front().size();
//
////    	papara_opencl pocl( ref_len, m_ref_pvecs.size() );
////
////    	pocl.setup_qs( m_qs_pvecs );
////    	pocl.run_refs(m_ref_pvecs, m_ref_aux );
//
////    	m_cl_res = pocl.get_res_map();
//
////    	papara_gpu_driver pdocl( m_ref_pvecs, m_ref_aux, m_qs_pvecs );
////
////    	pdocl.run();
////    	m_cl_res = pdocl.get_res_map();
//
//    	std::auto_ptr<papara_gpu_driver> pdocl( new_papara_gpu_driver(m_ref_pvecs, m_ref_aux, m_qs_pvecs));
//    	pdocl->run();
//    	m_cl_res = pdocl->get_res_map();
//    }
//
//
//    void run_seq() {
//        align_arrays arr;
//
//        uint64_t ncup = 0;
//
//        timer t1;
//
//        for( size_t i = 0; i < m_ref_pvecs.size(); ++i ) {
//
//            const size_t reflen = m_ref_pvecs[i].size();
//
//            for( size_t j = 0; j < m_qs_pvecs.size(); ++j ) {
//
//
//        //        int score2 = 0;
//                int score = align_pvec_score_seq( m_ref_pvecs[i], m_ref_aux[i], m_qs_pvecs[j], score_mismatch, score_match_cgap, score_gap_open, score_gap_extend, arr );
//                //std::cout << "score " << i << " " << j << " = " << score << " " << score2 << "\n";
////                 if( score != score2 ) {
//   //                 std::cout << "score " << i << " " << j << " = " << score << "\n";
////                 }
//                ncup += reflen * m_qs_pvecs[i].size();
//
//                m_seq_res[std::pair<size_t,size_t>(i,j)] = score;
//            }
//
//
//        }
//
//        std::cerr << "aligned " << m_ref_pvecs.size() << " x " << m_qs_pvecs.size() << " seqs in " << t1.elapsed() << " s\n";
//        std::cerr << "GNcups: " << ncup / (t1.elapsed() * 1e9) << "\n";
//    }
//
//    template<typename score_t, const size_t W>
//    void run_vec() {
//        align_arrays arr;
//
//        std::cout << "run vec " << W << "\n";
//
//        uint64_t ncup = 0;
//
//        timer t1;
//
//
//        for( size_t i = 0; i < m_ref_pvecs.size(); i += W ) {
//            const int *seqptrs[W];
//            const unsigned int *auxptrs[W];
//            size_t num_valid = 0;
//            for( size_t j = 0; j < W; ++j ) {
//                if( i + j < m_ref_pvecs.size() ) {
//                    seqptrs[j] = m_ref_pvecs[i+j].data();
//                    auxptrs[j] = m_ref_aux[i+j].data();
//                    ++num_valid;
//                } else {
//                    seqptrs[j] = seqptrs[0];
//                    auxptrs[j] = auxptrs[0];
//                }
//            }
//
//            assert( num_valid > 0 );
//
//            const size_t reflen = m_ref_pvecs[i].size();
//
//            const align_pvec_score<score_t,W> aligner(seqptrs, auxptrs, reflen, score_mismatch, score_match_cgap, score_gap_open, score_gap_extend );
//
//
////             std::cerr << "ref: " << i << "\n";
//            for( size_t j = 0; j < m_qs_pvecs.size(); ++j ) {
//
//                aligner.align( m_qs_pvecs[j] );
//                const score_t *scores = aligner.get_scores();
//
//                for( size_t k = 0; k < num_valid; ++k ) {
//
//           //         std::cout << "score " << i + k << " " << j << " = " << scores[k] << "\n";
//                    m_vec_res[std::pair<size_t,size_t>(i+k,j)] = scores[k];
//                }
//                ncup += num_valid * reflen * m_qs_pvecs[i].size();
//            }
//
//
//        }
//
//        std::cerr << "aligned " << m_ref_pvecs.size() << " x " << m_qs_pvecs.size() << " seqs in " << t1.elapsed() << " s\n";
//        std::cerr << "GNcups: " << ncup / (t1.elapsed() * 1e9) << "\n";
//    }
//
//
//    void compare() {
//    	std::cout << "\n";
//        for( std::map< std::pair< size_t, size_t >, int >::iterator it = m_seq_res.begin(); it != m_seq_res.end(); ++it ) {
//            int seq_score = it->second;
//            int vec_score = m_vec_res[it->first];
//
//            if( seq_score != vec_score ) {
//                std::cerr << "error: " << seq_score << " != " << vec_score << "\n";
//            }
//
//            std::cout << seq_score << " ";
//
//        }
//        std::cout << "checked\n";
//
//    }
//
//    void compare2() {
//    	std::cout << "\n";
//        for( std::map< std::pair< size_t, size_t >, int >::iterator it = m_vec_res.begin(); it != m_vec_res.end(); ++it ) {
//            int ref_score = it->second;
//            int cl_score = m_cl_res[it->first];
//
//            if( ref_score != cl_score ) {
//                std::cerr << it->first.first << " " << it->first.second << ": error: " << ref_score << " != " << cl_score << "\n";
//            }
//
////            std::cout << ref_score << " ";
//
//        }
//        std::cout << "checked\n";
//
//    }
//};


#if 0
int main( int argc, char *argv[] ) {

	size_t num_refs = atoi(argv[1] );
	size_t num_qs = atoi(argv[2] );



    testbench tb( "qs.bin", "ref.bin", num_refs, num_qs );
//    tb.run_vec<short,8>();
//    tb.run_vec<int,4>();



    assert( num_refs == 128 );
    //assert( num_qs == 20 );



    tb.run_opencl();

    tb.run_vec<int,4>();

    tb.compare2();

	//tb.run_seq();



//    if( true ) {
//    	tb.run_seq();
//
//    	tb.compare();
//    }
}
#endif
